package com.emsoft.collections;


import com.emsoft.common.Assert;
import com.emsoft.observer.ObservableImpl;
import com.emsoft.observer.Observer;

import java.util.*;


/**
 * Observable implementation of List. Registers subscribers and notify their about changes of elements.
 *
 * @param <E>
 * @author Vladimir Zhukov
 */
public class ObservableList<E> extends ObservableImpl<Observer<ListEditEventArgs>, ListEditEventArgs>
        implements List<E> {

    private List<E> list;

    public ObservableList() {
        this(new ArrayList<E>());
    }

    public ObservableList(List<E> list) {
        Assert.notNull(list, "list");
        this.list = list;
    }

    @Override
    public boolean add(E e) {
        boolean success = list.add(e);
        if (success)
            notifyObservers(new ListEditEventArgs(ListEditAction.INSERT, -1));
        return success;
    }

    @Override
    public void add(int index, E element) {
        list.add(index, element);
        notifyObservers(new ListEditEventArgs(ListEditAction.INSERT, index));
    }

    @Override
    public boolean addAll(Collection<? extends E> c) {
        boolean success = list.addAll(c);
        if (success)
            notifyObservers(new ListEditEventArgs(ListEditAction.MULTI_INSERT, -1));
        return success;
    }

    @Override
    public boolean addAll(int index, Collection<? extends E> c) {
        boolean success = list.addAll(index, c);
        if (success)
            notifyObservers(new ListEditEventArgs(ListEditAction.MULTI_INSERT, -1));
        return success;
    }

    @Override
    public boolean remove(Object o) {
        int index = list.indexOf(o);
        boolean success = list.remove(o);
        if (success)
            notifyObservers(new ListEditEventArgs(ListEditAction.DELETE, index));
        return success;
    }

    @Override
    public E remove(int index) {
        E el = list.remove(index);
        notifyObservers(new ListEditEventArgs(ListEditAction.DELETE, index));
        return el;
    }

    @Override
    public boolean removeAll(Collection<?> c) {
        boolean success = list.removeAll(c);
        if (success)
            notifyObservers(new ListEditEventArgs(ListEditAction.MULTI_DELETE, -1));
        return success;
    }

    @Override
    public boolean retainAll(Collection<?> c) {
        boolean success = list.retainAll(c);
        if (success)
            notifyObservers(new ListEditEventArgs(ListEditAction.MULTI_DELETE, -1));
        return success;
    }

    @Override
    public void clear() {
        int count = list.size();
        list.clear();
        if (count > 0)
            notifyObservers(new ListEditEventArgs(ListEditAction.CLEAR, -1));
    }

    @Override
    public E set(int index, E element) {
        E oldValue = list.set(index, element);
        notifyObservers(new ListEditEventArgs(ListEditAction.UPDATE, index));
        return oldValue;
    }

    @Override
    public Iterator<E> iterator() {
        return new ObservableIterator(list.iterator());
    }

    @Override
    public ListIterator<E> listIterator() {
        return new ObservableListIterator(list.listIterator(), 0);
    }

    @Override
    public ListIterator<E> listIterator(int index) {
        return list.listIterator(index);
    }

    @Override
    public boolean containsAll(Collection<?> c) {
        return list.containsAll(c);
    }

    @Override
    public E get(int index) {
        return list.get(index);
    }

    @Override
    public int size() {
        return list.size();
    }

    @Override
    public boolean isEmpty() {
        return list.isEmpty();
    }

    @Override
    public boolean contains(Object o) {
        return list.contains(o);
    }

    @Override
    public Object[] toArray() {
        return list.toArray();
    }

    @Override
    public <T> T[] toArray(T[] a) {
        return list.toArray(a);
    }

    @Override
    public int indexOf(Object o) {
        return list.indexOf(o);
    }

    @Override
    public int lastIndexOf(Object o) {
        return list.lastIndexOf(o);
    }

    @Override
    public List<E> subList(int fromIndex, int toIndex) {
        return list.subList(fromIndex, toIndex);
    }

    private class ObservableIterator implements Iterator<E> {

        private int cursor;
        private final Iterator<E> delegate;

        public ObservableIterator(Iterator<E> iterDelegate) {
            this.delegate = iterDelegate;
            cursor = 0;
        }

        protected Iterator<E> getDelegate() {
            return delegate;
        }

        @Override
        public boolean hasNext() {
            return delegate.hasNext();
        }

        @Override
        public E next() {
            cursor++;
            return delegate.next();
        }

        @Override
        public void remove() {
            ObservableList.this.remove(cursor--);
        }

        public int getCursor() {
            return cursor;
        }

        public void setCursor(int cursor) {
            this.cursor = cursor;
        }
    }

    private class ObservableListIterator extends ObservableIterator implements ListIterator<E> {

        public ObservableListIterator(ListIterator<E> iterDelegate, int index) {
            super(iterDelegate);
            setCursor(index);
        }

        public ListIterator<E> getListIterator() {
            return (ListIterator<E>)getDelegate();
        }

        @Override
        public void add(E o) {
            ObservableList.this.add(o);
            setCursor(getCursor() + 1);
        }

        @Override
        public boolean hasPrevious() {
            return getListIterator().hasPrevious();
        }

        @Override
        public int nextIndex() {
            return getListIterator().nextIndex();
        }

        @Override
        public E previous() {
            return getListIterator().previous();
        }

        @Override
        public E next() {
            return getListIterator().next();
        }

        @Override
        public int previousIndex() {
            return getListIterator().previousIndex();
        }

        @Override
        public void set(E o) {
            ObservableList.this.set(getCursor(), o);
        }
    }
}