package com.emsoft.exceptions

import com.emsoft.common.Assert
import com.emsoft.common.ResourceHelper
import com.emsoft.common.StringHelper
import org.slf4j.LoggerFactory

/**
 * Utils for exceptions.
 *
 * @author Vladimir Zhukov
 */
class ExHelper {

    /**
     * Resource file of error messages.
     */
    final static String ERROR_MESSAGES_RESOURCE = "com/emsoft/utils/resources/ErrorMessages"
    final static String DEFAULT_LOGGER = "default_error_logger"

    protected ExHelper() {
    }

    /**
     * Returns text description of error from resources.
     *
     * @param key
     * @return
     */
    private static String getLocalMessage(String key) {
        ResourceBundle rb = ResourceHelper.getResourceBundleSafe(ERROR_MESSAGES_RESOURCE)
        if (rb == null)
            throw new InternalException(ErrorCode.INTERNAL_ERROR,
                    """Resource '${ERROR_MESSAGES_RESOURCE}.${ResourceHelper.BUNDLE_EXTENSION}' for error messages not found.""")
        return rb.getString(key)
    }

    /**
     * Returns text description of error.
     *
     * @param ex
     * @param includeInnerException
     * @param includeStackTrace
     * @return
     */
    static String getMessage(Throwable ex, boolean includeInnerException = true,
            boolean includeStackTrace = false) {
        Assert.notNull(ex, "ex")
        StringBuilder sb = new StringBuilder()
        appendMessage(sb, ex)
        if (includeInnerException) {
            Throwable innerEx = ex.cause
            def tab = "  "
            while (innerEx != null) {
                appendMessage(sb, innerEx, tab)
                innerEx = innerEx.cause
                tab += "  "
            }
        }
        if (includeStackTrace)
            fillStackTrace(sb, ex, 0)
        return sb.toString()
    }

    private static void appendMessage(StringBuilder sb, Throwable ex, String tab = StringHelper.EMPTY) {
        if (sb.length() > 0) {
            sb.append(StringHelper.CR)
            sb.append(tab)
        }
        sb.append(ex.class.simpleName)
        sb.append(": ")
        sb.append(ex.localizedMessage)
    }

    /**
     * Returns stacktrace.
     *
     * @param ex
     * @return
     */
    static String getStackTrace(Throwable ex, int stackTraceDepth = 0) {
        Assert.notNull(ex, "ex")
        Assert.notNull(stackTraceDepth, "stackTraceDepth")
        def sb = new StringBuilder()
        fillStackTrace(sb, ex, stackTraceDepth)
        return sb.toString()
    }

    private static fillStackTrace(StringBuilder sb, Throwable ex, int stackTraceDepth) {
        StackTraceElement[] ar = ex.stackTrace
        if (ar) {
            if (sb.length() > 0)
                sb.append(StringHelper.CR)
            sb.append("StackTrace: ")
            int depth = (stackTraceDepth > 0 ? Math.min(stackTraceDepth, ar.length) : ar.length)
            def tab = StringHelper.EMPTY
            for (i in 0..depth - 1) {
                if (i > 0) {
                    sb.append(StringHelper.CR)
                    sb.append(tab)
                }
                sb.append(ar[i].toString())
                tab += "  "
            }
        }
    }

    /**
     * Returns first element of stack trace.
     *
     * @param ex
     * @return
     */
    static StackTraceElement getRootStackTrace(Throwable ex) {
        Assert.notNull(ex, "ex")
        StackTraceElement[] stAr = ex.stackTrace
        if (stAr)
            return stAr[stAr.length - 1]
        return null
    }

    /**
     * Returns class name of the first element of stack trace in which the exception occurred.
     *
     * @param ex
     * @return
     */
    static String getRootClassName(Throwable ex) {
        StackTraceElement trace = getRootStackTrace(ex)
        if (trace == null)
            return StringHelper.EMPTY
        return trace.getClassName()
    }

    static void throwException(String messageCode, ErrorCode errorCode, Object... params) {
        throw new InternalException(errorCode, String.format(getLocalMessage(messageCode), params))
    }

    static void throwException(String messageCode, Object... params) {
        throwException(messageCode, ErrorCode.INTERNAL_ERROR, params)
    }

    /**
     * Logged exception with slf4j, wrap in BaseRuntimeException if it not instance of RuntimeException and throw.
     *
     * @param ex
     * @param className
     */
    static void logAndThrow(Throwable ex, String className = getRootClassName(ex)) {
        Assert.hasText(className, "className")
        LoggerFactory.getLogger(className).error(getMessage(ex), ex)
        def runtimeEx = ex instanceof RuntimeException ? ex : new InternalException(ErrorCode.INTERNAL_ERROR, ex)
        throw runtimeEx
    }

    static void throwNullArgument(String name) {
        throw new IllegalArgumentException(String.format(
                getLocalMessage("NullArgumentExceptionMessage"), name))
    }

    static void throwStringArgumentNullOrEmpty(String name) {
        throw new IllegalArgumentException(String.format(
                getLocalMessage("StringArgumentNullOrEmptyMessage"), name))
    }

    static void throwValueAlreadyExists(Object value) {
        throwException("ValueAlreadyExistsMessage", [value])
    }

    static void throwObjectNotInCollection(String name) {
        throw new ObjectNotInCollectionException(String.format(
                getLocalMessage("ObjectNotInCollectionExceptionMessage"), name))
    }

    static void throwIndexOutOfBounds() {
        throw new IndexOutOfBoundsException(
                getLocalMessage("ObjectNotInCollectionExceptionMessage"))
    }

    static void throwEmptyCollection(String col) {
        throw new EmptyCollectionException(col)
    }

    static void throwUnsupportedOperation() {
        throw new UnsupportedOperationException(
                getLocalMessage("UnsupportedOperationExceptionMessage"))
    }

    static void throwUnknownType(String type) {
        throw new InternalException(ErrorCode.INTERNAL_ERROR, String.format(
                getLocalMessage("UnknownTypeExceptionMessage"), type))
    }

    static void throwResourceNotFound(String bundleName) {
        throw new ResourceNotFoundException(String.format(
                getLocalMessage("ResourceNotFoundExceptionMessage"), bundleName))
    }

    static void throwPropertyNotInitialized(String name) {
        throw new InternalException(ErrorCode.INTERNAL_ERROR, String.format(
                getLocalMessage("PropertyNotInitializedExceptionMessage"), name))
    }
}
