package com.emsoft.security

import com.emsoft.common.Assert
import com.emsoft.common.StringHelper
import org.codehaus.groovy.runtime.EncodingGroovyMethods

/**
 * Class for work with password.
 *
 * @author Vladimir Zhukov
 */
class Password {

	private String algorithm
    /**
     * saltBytes (synchrosignal)
     */
    private byte[] saltBytes
    /**
     * hashBytes of password
     */
    private byte[] hashBytes

    /**
     * @param hash String in Base64 encoding.
     * @param salt String in Base64 encoding.
     * @return
     */
    Password(String hash, String salt, String algorithm = HashHelper.SHA256) {
        Assert.hasText(hash, "hash")
		Assert.hasText(algorithm, "algorithm")
        this.hashBytes = EncodingGroovyMethods.decodeBase64(hash)
        if (salt)
            this.saltBytes = EncodingGroovyMethods.decodeBase64(salt)
		this.algorithm = algorithm
    }

    Password(byte[] hash, byte[] salt, String algorithm = HashHelper.SHA256) {
        Assert.notNull(hash, "hash")
		Assert.hasText(algorithm, "algorithm")
        this.hashBytes = hash.clone()
        if (salt)
            this.saltBytes = salt.clone()
		this.algorithm = algorithm
    }

    /**
     * @param clearText Password string.
     * @param useSalt If true then generates random salt.
     * @return
     */
    Password(String clearText, boolean useSalt, String algorithm = HashHelper.SHA256) {
        Assert.hasText(clearText, "clearText")
		Assert.hasText(algorithm, "algorithm")
        if (useSalt)
            this.saltBytes = HashHelper.generateRandomBytes(6)
        this.hashBytes = HashHelper.hash(clearText, saltBytes)
		this.algorithm = algorithm
    }

    /**
     *
     * @param clearText Password char array.
     * @param useSalt If true then generates random salt.
     * @return
     */
    Password(char[] clearText, boolean useSalt) {
        Assert.hasElements(clearText, "clearText")
        if (useSalt)
            this.saltBytes = HashHelper.generateRandomBytes(6)
        this.hashBytes = HashHelper.hash(clearText, saltBytes)
    }

    /**
     * @return String of salt in Base64.
     */
    String getSalt() {
        return (saltBytes ? EncodingGroovyMethods.encodeBase64(saltBytes).toString() : null)
    }

    /**
     * @return String of hash in Base64.
     */
    String getHash() {
        return EncodingGroovyMethods.encodeBase64(hashBytes).toString()
    }

    byte[] getRawSalt() {
        if (saltBytes)
            return saltBytes.clone()
        return null
    }

    byte[] getRawHash() {
        return hashBytes.clone()
    }

    /**
     * Compares current hash with hash of clearText.
     *
     * @param clearText
     * @return
     */
    boolean verify(String clearText) {
        Assert.hasText(clearText, "clearText")
        return HashHelper.verify(clearText, saltBytes)
    }

    /**
     * Compares current hash with hash of clearText.
     *
     * @param clearText
     * @return
     */
    boolean verify(char[] clearText) {
        Assert.hasElements(clearText, "clearText")
        return HashHelper.verify(clearText, saltBytes)
    }

    /**
     * Generates random password.
     *
     * @return
     */
    static String generatePassword(int bytesLength = 9) {
        // generate 9 random bytes
        // this is enough to get 12 random characters from the set of base64
        byte[] rnd = HashHelper.generateRandomBytes(bytesLength)
        // convert random bytes in base64
        String password = EncodingGroovyMethods.encodeBase64(rnd).toString()
        // clear the workin array
        Arrays.fill(rnd, (Byte) 0)
        return password
    }
}